# Standard library imports
import os

from .cleanup import extract_highest_accuracy_model
from .prepare_data import split_training_2_validation
from .run_training import apply_feedback, manage_fine_tuning_config, run_main_train_code


def train(
    input_path: str,
    output_path: str,
    epoch_size: int,
    batch_size: int,
    model: str,
    model_home: str,
    freeze_layers: bool = False,
):
    """Trains the input image with base model


    Args:
        input_path: Path of the directory output by preprocess
        output_path: Path of the working dir for training
        epoch_size: Epoch size to be used for training
        batch_size: Batch size to be used for training
        model : Choose Model, Options supported are , ramp
        model_home : Model Home directory which contains necessary file in order to run model
    Example::

        final_accuracy, final_model_path = train(
            input_path=preprocess_output,
            epoch_size=2,
            batch_size=2,
            model="ramp",
            model_home=os.environ["RAMP_HOME"],
        )
    """
    assert os.path.exists(input_path), "Input Path Doesn't Exist"
    assert os.path.exists(model_home), "Model Home Doesn't Exist"
    supported_models = ["ramp"]

    # Use the assert keyword to raise an AssertionError if the input model string is not in the list of supported models
    assert any(
        model.lower() in supported_model for supported_model in supported_models
    ), "Model is not in the list of supported models "

    # Export the environment variables from the operating system
    os.environ.update(os.environ)
    if model.lower() == "ramp":
        # Add a new environment variable to the operating system
        os.environ["RAMP_HOME"] = model_home
        # Print the environment variables to verify that the new variable was added
        print("Starting to prepare data for training")
        split_training_2_validation(input_path, output_path)
        cfg = manage_fine_tuning_config(
            output_path, epoch_size, batch_size, freeze_layers
        )
        print("Data is ready for training")
        run_main_train_code(cfg)
        print("extracting highest accuracy model")
        final_accuracy, final_model_path = extract_highest_accuracy_model(output_path)
        return (final_accuracy, final_model_path)


def run_feedback(
    input_path,
    output_path,
    feedback_base_model,
    model_home: str,
    epoch_size: int,
    batch_size: int,
    freeze_layers: bool = True,
):
    assert os.path.exists(input_path), "Input Feedback Path Doesn't Exist"
    assert os.path.exists(feedback_base_model), "Feedback base Model Doesn't Exist"
    os.environ.update(os.environ)
    os.environ["RAMP_HOME"] = model_home
    print("Starting to prepare data for training")
    split_training_2_validation(input_path, output_path)
    print("Data is ready for training")

    apply_feedback(
        feedback_base_model, output_path, epoch_size, batch_size, freeze_layers
    )
    final_accuracy, final_model_path = extract_highest_accuracy_model(output_path)
    return (final_accuracy, final_model_path)



# Standard library imports
import os

from ..georeferencing import georeference
from .clip_labels import clip_labels
from .fix_labels import fix_labels
from .reproject_labels import reproject_labels_to_epsg3857


def preprocess(
    input_path: str,
    output_path: str,
    rasterize=False,
    rasterize_options=None,
    georeference_images=False,
) -> None:
    """Fully preprocess the input data.

    The preprocessed images and labels are in the EPSG:3857 projected
    coordinate system ('WGS 84 / Pseudo-Mercator', coordinates
    in meters).

    Args:
        input_path: Path of the directory with input data.
        output_path: Path of the directory where to put the preprocessed
            data. New directories will be created in this directory:
            "labels" - for the clipped GeoJSON labels,
            "chips" - for the georeferenced OAM images
            (if georeference_images=True), and the directories
            "binarymasks" and "grayscale_labels" if the corresponding
            rasterizing options are chosen.
        rasterize: Whether to create the raster labels.
        rasterize_options: A list with options how to rasterize the
            label, if rasterize=True. Possible options: "grayscale"
            (burn value will be 255, white buildings on black
            background), "binary" (burn value will be 1, can be used
            for the ramp model).
            If rasterize=False, rasterize_options will be ignored.
        georeference_images: Whether to georeference the OAM images.

    Example::

        preprocess(
            "data/region1_inputs",
            "data/region1_preprocessed",
            rasterize=True,
            rasterize_options=["grayscale", "binary"],
            georeference_images=True
        )
    """
    # Check if rasterizing options are valid
    if rasterize:
        assert (
            rasterize_options is not None
            and isinstance(rasterize_options, list)
            and 0 < len(rasterize_options) <= 2
            and len(rasterize_options) == len(set(rasterize_options))
        ), "Please provide a list with rasterizing options"

        for option in rasterize_options:
            assert option in (
                "grayscale",
                "binary",
            ), "Please provide a list with valid rasterizing options"

    os.makedirs(output_path, exist_ok=True)

    if georeference_images:
        georeference(input_path, f"{output_path}/chips")

    fix_labels(
        f"{input_path}/labels.geojson",
        f"{output_path}/corrected_labels.geojson",
    )

    reproject_labels_to_epsg3857(
        f"{output_path}/corrected_labels.geojson",
        f"{output_path}/labels_epsg3857.geojson",
    )

    clip_labels(input_path, output_path, rasterize, rasterize_options)

    os.remove(f"{output_path}/corrected_labels.geojson")
    os.remove(f"{output_path}/labels_epsg3857.geojson")



# Standard library imports
import concurrent.futures
import io
import json
import math
import os
import re
import time
import urllib.request
import zipfile
from glob import glob
from typing import Tuple

# Third party imports
# Third-party imports
import geopandas
import requests
from shapely.geometry import box

IMAGE_SIZE = 256


def get_prefix(path: str) -> str:
    """Get filename prefix (without extension) from full path."""
    filename = os.path.basename(path)
    return os.path.splitext(filename)[0]


def get_bounding_box(filename: str) -> Tuple[float, float, float, float]:
    """Get the EPSG:3857 coordinates of bounding box for the OAM image.

    This function gives the coordinates of lower left and upper right
    corners of the OAM image. We will use the bounding box to georeference
    the image and for clipping and rasterizing the corresponding labels.

    Returns:
        A tuple, (x_min, y_min, x_max, y_max), with coordinates in meters.
    """
    filename = re.sub(r'\.(png|jpeg)$', '', filename)
    _, *tile_info = re.split("-", filename)
    x_tile, y_tile, zoom = map(int, tile_info)

    # Lower left and upper right corners in degrees
    x_min, y_min = num2deg(x_tile, y_tile + 1, zoom)
    x_max, y_max = num2deg(x_tile + 1, y_tile, zoom)

    # Create a GeoDataFrame containing a polygon for bounding box
    box_4326 = box(x_min, y_min, x_max, y_max)
    gdf_4326 = geopandas.GeoDataFrame({"geometry": [box_4326]}, crs="EPSG:4326")

    # Reproject to EPSG:3857
    gdf_3857 = gdf_4326.to_crs("EPSG:3857")

    # Bounding box in EPSG:3857 as a tuple (x_min, y_min, x_max, y_max)
    box_3857 = gdf_3857.iloc[0, 0].bounds

    return box_3857


def num2deg(x_tile: int, y_tile: int, zoom: int) -> Tuple[float, float]:
    """Convert tile numbers to EPSG:4326 coordinates.

    Convert tile numbers to the WGS84 longitude/latitude coordinates
    (in degrees) of the upper left corner of the tile.

    Args:
        x_tile: Tile X coordinate
        y_tile: Tile Y coordinate
        zoom: Level of detail

    Returns:
        A tuple (longitude, latitude) in degrees.
    """
    n = 2.0**zoom
    lon_deg = x_tile / n * 360.0 - 180.0
    lat_rad = math.atan(math.sinh(math.pi * (1 - 2 * y_tile / n)))
    lat_deg = math.degrees(lat_rad)

    return lon_deg, lat_deg


def remove_files(pattern: str) -> None:
    """Remove files matching a wildcard."""
    files = glob(pattern)
    for file in files:
        os.remove(file)


def convert2worldcd(lat, lng, tile_size):
    """
    World coordinates  are measured from the Mercator projection's origin
    (the northwest corner of the map at 180 degrees longitude and
    approximately 85 degrees latitude) and increase in the x direction
    towards the east (right) and increase in the y direction towards the south
    (down).Because the basic Mercator  tile is 256 x 256 pixels, the usable
    world coordinate space is {0-256}, {0-256}
    """
    siny = math.sin((lat * math.pi) / 180)
    siny = min(max(siny, -0.9999), 0.9999)
    world_x = tile_size * (0.5 + (lng / 360))
    world_y = tile_size * (0.5 - math.log((1 + siny) / (1 - siny)) / (4 * math.pi))
    # print("world coordinate space is %s, %s",world_x,world_y)
    return world_x, world_y


def latlng2tile(zoom, lat, lng, tile_size):
    """By dividing the pixel coordinates by the tile size and taking the
    integer parts of the result, you produce as a by-product the tile
    coordinate at the current zoom level."""
    zoom_byte = 1 << zoom  # converting zoom level to pixel bytes
    # print(zoom_byte)
    w_x, w_y = convert2worldcd(lat, lng, tile_size)

    t_x = math.floor((w_x * zoom_byte) / tile_size)
    t_y = math.floor((w_y * zoom_byte) / tile_size)
    return t_x, t_y


def bbox2tiles(bbox_coords, zm_level, tile_size):
    # start point where we will start downloading the tiles

    start_point_lng = bbox_coords[0]  # getting the starting lat lng
    start_point_lat = bbox_coords[1]

    # end point where we should stop downloading the tile
    end_point_lng = bbox_coords[2]  # getting the ending lat lng
    end_point_lat = bbox_coords[3]

    # Note :  lat=y-axis, lng=x-axis
    # getting tile coordinate for first point of bbox
    start_x, start_y = latlng2tile(
        zoom=zm_level,
        lat=start_point_lat,
        lng=start_point_lng,
        tile_size=tile_size,
    )
    start = [start_x, start_y]

    # getting tile coordinate for last point of bbox
    end_x, end_y = latlng2tile(
        zoom=zm_level,
        lat=end_point_lat,
        lng=end_point_lng,
        tile_size=tile_size,
    )
    end = [end_x, end_y]
    return start, end


def download_image(url, base_path, source_name):
    response = requests.get(url)
    image = response.content

    url_splitted_list = url.split("/")
    filename = f"{base_path}/{source_name}-{url_splitted_list[-2]}-{url_splitted_list[-1]}-{url_splitted_list[-3]}.png"

    with open(filename, "wb") as f:
        f.write(image)

    # print(f"Downloaded: {url}")


def tms2img(start: list, end: list, zm_level, base_path, source="maxar"):
    """Downloads imagery from start to end tile coordinate system

    Args:
        start (list):[tile_x,tile_y]
        end (list): [tile_x,tile_y],
        source (string): it should be eithre url string or maxar value
        zm_level : Zoom level
        base_path : Source where image will be downloaded

    """

    begin_x = start[0]  # this will be the beginning of the download loop for x
    begin_y = start[1]  # this will be the beginning of the download loop for x
    stop_x = end[0]  # this will be the end of the download loop for x
    stop_y = end[1]  # this will be the end of the download loop for x

    print(f"Download starting from {start} to {end} using source {source} - {zm_level}")

    start_x = begin_x  # starting loop from beginning
    start_y = begin_y  # starting y loop from beginnig
    source_name = "OAM"  # default
    download_urls = []
    while start_x <= stop_x:  # download  x section while keeping y as c
        start_y = begin_y
        while start_y >= stop_y:  # download  y section while keeping x as c
            download_path = [start_x, start_y]
            if source == "maxar":
                try:
                    connect_id = os.environ.get("MAXAR_CONNECT_ID")
                except Exception as ex:
                    raise ex
                source_name = source
                download_url = f"https://services.digitalglobe.com/earthservice/tmsaccess/tms/1.0.0/DigitalGlobe:ImageryTileService@EPSG:3857@jpg/{zm_level}/{download_path[0]}/{download_path[1]}.jpg?connectId={connect_id}&flipy=true"

            # add multiple logic on supported sources here
            else:
                # source should be url as string , like this :  https://tiles.openaerialmap.org/62dbd947d8499800053796ec/0/62dbd947d8499800053796ed/{z}/{x}/{y}
                download_url = source.format(
                    x=download_path[0], y=download_path[1], z=zm_level
                )
            download_urls.append(download_url)

            start_y = start_y - 1  # decrease the y

        start_x = start_x + 1  # increase the x

    # Use the ThreadPoolExecutor to download the images in parallel
    with concurrent.futures.ThreadPoolExecutor() as executor:
        for url in download_urls:
            executor.submit(download_image, url, base_path, source_name)


def fetch_osm_data(payload: json, API_URL="https://raw-data-api0.hotosm.org/v1"):
    """
    args :
        payload : Payload request for API URL
        API_URL : Raw data API URL
    Returns :
        geojson
    """
    headers = {"accept": "application/json", "Content-Type": "application/json"}

    task_response = requests.post(
        url=f"{API_URL}/snapshot/", data=json.dumps(payload), headers=headers
    )

    task_response.raise_for_status()
    result = task_response.json()
    print(result)
    task_track_url = result["track_link"]
    stop_loop = False
    while not stop_loop:
        check_result = requests.get(url=f"{API_URL}{task_track_url}")
        check_result.raise_for_status()
        res = (
            check_result.json()
        )  # status will tell current status of your task after it turns to success it will give result
        if res["status"] == "SUCCESS" or res["status"] == "FAILED":
            stop_loop = True
        time.sleep(1)  # check each second
    # Download the zip file from the URL
    url = res["result"]["download_url"]
    response = urllib.request.urlopen(url)

    # Open the zip file from the response data
    with zipfile.ZipFile(io.BytesIO(response.read()), "r") as zip_ref:
        with zip_ref.open("Export.geojson") as file:
            my_export_geojson = json.loads(file.read())
    return my_export_geojson



# Import the exec function
from __future__ import print_function

# Standard library imports
import os
import subprocess
import sys
from pathlib import Path
from shutil import copytree, rmtree


class RaiseError(Exception):
    def __init__(self, message):
        self.message = message


def split_training_2_validation(input_path, output_path):
    """Converts training 2 validation

    Currently supported for ramp , It converts training dataset provided by preprocessing script to validation datastes reuqired by ramp
    """

    RAMP_HOME = os.environ["RAMP_HOME"]
    PYTHON_HOME = os.environ.get("PYTHON_HOME")
    python_exec = "python"
    if PYTHON_HOME:
        python_exec = PYTHON_HOME

    sys.path.append("..")
    os.chdir(Path(RAMP_HOME))
    # output current working directory.

    # Define the source and destination paths
    src_path = input_path
    dst_path = output_path

    # Check if the path exists
    if os.path.exists(dst_path):
        # Delete the directory and its contents
        rmtree(dst_path)
    # Use the copytree function to copy the source directory and its contents to the destination directory
    copytree(src_path, dst_path)

    # Define the script as a string
    # SPLIT INTO TRAINING AND VALIDATION
    # script = f"""%run ramp-code/scripts/make_train_val_split_lists.py -src {dst_path}/chips -pfx {uid}_fair_split -trn 0.85 -val 0.15"""
    try:
        subprocess.check_output(
            [
                python_exec,
                f"{RAMP_HOME}/ramp-code/scripts/make_train_val_split_lists.py",
                "-src",
                f"{dst_path}/chips",
                "-pfx",
                f"{dst_path}/fair_split",
                "-trn",
                "0.85",
                "-val",
                "0.15",
            ],
            env=os.environ,
        )
    except subprocess.CalledProcessError as ex:
        raise ex

    # move all the VALIDATION chips, labels and masks to their new locations
    try:
        subprocess.check_output(
            [
                python_exec,
                f"{RAMP_HOME}/ramp-code/scripts/move_chips_from_csv.py",
                "-sd",
                f"{dst_path}/chips",
                "-td",
                f"{dst_path}/val-chips",
                "-csv",
                f"{dst_path}/fair_split_val.csv",
                "-mv",
            ],
            env=os.environ,
        )
    except subprocess.CalledProcessError as ex:
        raise ex

    try:
        subprocess.check_output(
            [
                python_exec,
                f"{RAMP_HOME}/ramp-code/scripts/move_chips_from_csv.py",
                "-sd",
                f"{dst_path}/labels",
                "-td",
                f"{dst_path}/val-labels",
                "-csv",
                f"{dst_path}/fair_split_val.csv",
                "-mv",
            ],
            env=os.environ,
        )
    except subprocess.CalledProcessError as ex:
        raise ex

    try:
        subprocess.check_output(
            [
                python_exec,
                f"{RAMP_HOME}/ramp-code/scripts/move_chips_from_csv.py",
                "-sd",
                f"{dst_path}/binarymasks",
                "-td",
                f"{dst_path}/val-binarymasks",
                "-csv",
                f"{dst_path}/fair_split_val.csv",
                "-mv",
            ],
            env=os.environ,
        )
    except subprocess.CalledProcessError as ex:
        raise ex



from .georeferencing import georeference
from .inference import predict
from .postprocessing import polygonize, vectorize
from .preprocessing import preprocess
from .training import train
from .utils import bbox2tiles, tms2img



# Standard library imports
import os
from glob import glob
from pathlib import Path

# Third-party imports
import geopandas
from osgeo import gdal
from shapely.geometry import box
from tqdm import tqdm

from ..utils import get_bounding_box


def clip_labels(
    input_path: str, output_path: str, rasterize=False, rasterize_options=None
) -> None:
    """Clip and rasterize the GeoJSON labels for each aerial image.

    For each of the OAM images, the corresponding GeoJSON files are
    clipped first and saved in the "labels" directory.
    Then, the clipped GeoJSON files are converted to GeoTIFF files
    and saved in the "binarymasks" or "grayscale_labels"
    directories if the corresponding rasterizing options are chosen.

    The EPSG:3857 projected coordinate system is used
    ('WGS 84 / Pseudo-Mercator', coordinates in meters).

    Args:
        input_path: Path of the directory with input data.
        output_path: Path of the directory where the directories with
            output data will be created.
        rasterize: Whether to rasterize the clipped labels.
        rasterize_options: A list with options how to rasterize the
            label, if rasterize=True. Possible options: "grayscale"
            (burn value will be 255, white buildings on black
            background), "binary" (burn value will be 1, can be used
            for the ramp model).
            If rasterize=False, rasterize_options will be ignored.
    """
    # Check if rasterizing options are valid and create the directories
    # for the output
    if rasterize:
        assert (
            rasterize_options is not None
            and isinstance(rasterize_options, list)
            and 0 < len(rasterize_options) <= 2
            and len(rasterize_options) == len(set(rasterize_options))
        ), "Please provide a list with rasterizing options"

        for option in rasterize_options:
            assert option in (
                "grayscale",
                "binary",
            ), "Please provide a list with valid rasterizing options"

            if option == "grayscale":
                output_grayscale_path = f"{output_path}/grayscale_labels"
                os.makedirs(output_grayscale_path, exist_ok=True)

            if option == "binary":
                output_binary_path = f"{output_path}/binarymasks"
                os.makedirs(output_binary_path, exist_ok=True)

    # Create a directory where the clipped GeoJSON labels will be placed
    output_geojson_path = f"{output_path}/labels"
    os.makedirs(output_geojson_path, exist_ok=True)

    # Clipping GeoJSON labels
    for path in tqdm(
        glob(f"{input_path}/*.png"), desc=f"Clipping labels for {Path(input_path).stem}"
    ):
        filename = Path(path).stem
        geojson_file_all_labels = f"{output_path}/labels_epsg3857.geojson"
        clipped_geojson_file = f"{output_geojson_path}/{filename}.geojson"

        # Bounding box as a tuple
        x_min, y_min, x_max, y_max = get_bounding_box(filename)
        # Bounding box as a polygon
        bounding_box_polygon = box(x_min, y_min, x_max, y_max)

        # Read all labels into a GeoDataFrame, clip it and
        # write to GeoJSON
        gdf_all_labels = geopandas.read_file(geojson_file_all_labels)
        gdf_clipped = gdf_all_labels.clip(bounding_box_polygon)
        if len(gdf_clipped) > 0:
            gdf_clipped.to_file(clipped_geojson_file)
        else:
            schema = {"geometry": "Polygon", "properties": {"id": "int"}}
            crs = "EPSG:3857"
            gdf_clipped.to_file(clipped_geojson_file, schema=schema, crs=crs)

        # Rasterizing
        if rasterize:
            for option in rasterize_options:
                if option == "grayscale":
                    raster_file = f"{output_grayscale_path}/{filename}.tif"
                    burn_values = [255]

                if option == "binary":
                    raster_file = f"{output_binary_path}/{filename}.mask.tif"
                    burn_values = [1]

                # Rasterize clipped labels
                _ = gdal.Rasterize(
                    destNameOrDestDS=raster_file,
                    srcDS=clipped_geojson_file,
                    format="GTiff",
                    outputType=gdal.GDT_Byte,
                    outputBounds=[x_min, y_min, x_max, y_max],
                    width=256,
                    height=256,
                    burnValues=burn_values,
                )
                # Close raster dataset
                _ = None



# Standard library imports
from glob import glob
from pathlib import Path

# Third party imports
import geopandas as gpd
import numpy as np
import rasterio as rio
from rasterio.features import shapes
from rasterio.merge import merge
from shapely.geometry import Polygon, shape
from tqdm import tqdm

TOLERANCE = 0.5
AREA_THRESHOLD = 5


def vectorize(input_path: str, output_path: str , tolerance: float = 0.5, area_threshold: float = 5) -> None:
    """Polygonize raster tiles from the input path.

    Note that as input, we are expecting GeoTIF images with EPSG:3857 as
    CRS here. CRS of the resulting GeoJSON file will be EPSG:4326.

    Args:
        input_path: Path of the directory where the TIF files are stored.
        output_path: Path of the output file.
        tolerance (float, optional): Tolerance parameter for simplifying polygons. Defaults to 0.5 m.
        area_threshold (float, optional): Threshold for filtering polygon areas. Defaults to 5 sqm.

    Example::

        vectorize("data/masks_v2/4", "labels.geojson", tolerance=0.5, area_threshold=5)
    """
    base_path = Path(output_path).parents[0]
    base_path.mkdir(exist_ok=True, parents=True)

    raster_paths = glob(f"{input_path}/*.tif")
    with rio.open(raster_paths[0]) as src:
        kwargs = src.meta.copy()

    rasters = [rio.open(path) for path in raster_paths]
    mosaic, output = merge(rasters)

    # Close raster files after merging
    for raster in rasters:
        raster.close()

    polygons = [shape(s) for s, _ in shapes(mosaic, transform=output)]

    areas = [poly.area for poly in polygons]
    max_area, median_area = np.max(areas), np.median(areas)
    polygons = [
        Polygon(poly.exterior.coords)
        for poly in polygons
        if poly.area != max_area
        and poly.area / median_area > area_threshold
    ]

    gs = gpd.GeoSeries(polygons, crs=kwargs["crs"]).simplify(tolerance)
    gs = remove_overlapping_polygons(gs)
    if gs.empty:
        raise ValueError("No Features Found")
    gs.to_crs("EPSG:4326").to_file(output_path)


def remove_overlapping_polygons(gs: gpd.GeoSeries) -> gpd.GeoSeries:
    to_remove = set()
    gs_sindex = gs.sindex

    for i, p in tqdm(gs.items()):
        possible_matches_index = list(gs_sindex.intersection(p.bounds))
        possible_matches = gs.iloc[possible_matches_index]
        precise_matches = possible_matches[possible_matches.overlaps(p)]

        for j, q in precise_matches.items():
            if i != j:
                if p.area < q.area:  # Compare the areas of the polygons
                    to_remove.add(i)
                else:
                    to_remove.add(j)

    return gs.drop(list(to_remove))


# Standard library imports
import os
import time
from glob import glob
from pathlib import Path

# Third party imports
import numpy as np
from tensorflow import keras

from ..georeferencing import georeference
from ..utils import remove_files
from .utils import open_images, save_mask

BATCH_SIZE = 8
IMAGE_SIZE = 256
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"


def predict(
    checkpoint_path: str, input_path: str, prediction_path: str, confidence: float = 0.5
) -> None:
    """Predict building footprints for aerial images given a model checkpoint.

    This function reads the model weights from the checkpoint path and outputs
    predictions in GeoTIF format. The input images have to be in PNG format.

    The predicted masks will be georeferenced with EPSG:3857 as CRS.

    Args:
        checkpoint_path: Path where the weights of the model can be found.
        input_path: Path of the directory where the images are stored.
        prediction_path: Path of the directory where the predicted images will go.
        confidence: Threshold probability for filtering out low-confidence predictions.

    Example::

        predict(
            "model_1_checkpt.tf",
            "data/inputs_v2/4",
            "data/predictions/4"
        )
    """
    start = time.time()
    print(f"Using : {checkpoint_path}")
    model = keras.models.load_model(checkpoint_path)
    print(f"It took {round(time.time()-start)} sec to load model")
    start = time.time()

    os.makedirs(prediction_path, exist_ok=True)
    image_paths = glob(f"{input_path}/*.png")

    for i in range((len(image_paths) + BATCH_SIZE - 1) // BATCH_SIZE):
        image_batch = image_paths[BATCH_SIZE * i : BATCH_SIZE * (i + 1)]
        images = open_images(image_batch)
        images = images.reshape(-1, IMAGE_SIZE, IMAGE_SIZE, 3)

        preds = model.predict(images)
        preds = np.argmax(preds, axis=-1)
        preds = np.expand_dims(preds, axis=-1)
        preds = np.where(
            preds > confidence, 1, 0
        )  # Filter out low confidence predictions

        for idx, path in enumerate(image_batch):
            save_mask(
                preds[idx],
                str(f"{prediction_path}/{Path(path).stem}.png"),
            )
    print(
        f"It took {round(time.time()-start)} sec to predict with {confidence} Confidence Threshold"
    )
    keras.backend.clear_session()
    del model
    start = time.time()

    georeference(prediction_path, prediction_path, is_mask=True)
    print(f"It took {round(time.time()-start)} sec to georeference")

    remove_files(f"{prediction_path}/*.xml")
    remove_files(f"{prediction_path}/*.png")



# Third party imports
import numpy as np
from PIL import Image
from tqdm import tqdm

from .building_footprint import BuildingExtract
from .utils import tiles_from_directory


def get_polygons(
    pred_masks_path, polygons_path, kernel_opening=20, simplify_threshold=0.01
):
    """Generate GeoJSON polygons from predicted masks.

    Args:
      pred_masks_path: directory where the predicted mask are saved
      polygons_path: path to GeoJSON file to store features in
      kernel_opening: the opening morphological operation's kernel size in pixel
      simplify_threshold: the simplification accuracy as max. percentage of the arc length, in [0, 1]

    """
    bldg_extract = BuildingExtract(kernel_opening, simplify_threshold)
    tiles = list(tiles_from_directory(pred_masks_path))

    for tile, path in tqdm(tiles, unit="mask"):
        mask = np.array(Image.open(path).convert("P"), dtype=np.uint8)
        bldg_extract.extract(tile, mask)

    bldg_extract.save(polygons_path)



# Standard library imports
import os

# Third party imports
import tensorflow as tf

# Assert that the version of the library is greater than or equal to 2.9.2
assert tf.__version__ <= "2.9.2"  # tested up to 2.9.2

# Standard library imports

# Standard library imports
import datetime
import json
import os
from pathlib import Path
from time import perf_counter

# Third party imports
import tensorflow as tf
from matplotlib import pyplot as plt
from tensorflow import keras

# Note: this suppresses warning and other less urgent messages,
# and only allows errors to be printed.
# Comment this out if you are having mysterious problems, so you can see all messages.
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"


class RaiseError(Exception):
    def __init__(self, message):
        self.message = message


# Third party imports

# Third party imports
import segmentation_models as sm
from ramp.data_mgmt.data_generator import (
    test_batches_from_gtiff_dirs,
    training_batches_from_gtiff_dirs,
)
from ramp.training import (
    callback_constructors,
    loss_constructors,
    metric_constructors,
    model_constructors,
    optimizer_constructors,
)

# import ramp dependencies.
from ramp.training.augmentation_constructors import get_augmentation_fn
from ramp.utils.misc_ramp_utils import get_num_files

# Segmentation Models: using `keras` framework.
sm.set_framework("tf.keras")


# this variable must be defined. It is the parent of the 'ramp-code' directory.
working_ramp_home = os.environ["RAMP_HOME"]


def apply_feedback(
    pretrained_model_path, output_path, num_epochs, batch_size, freeze_layers
):
    if not os.path.exists(output_path):
        os.makedirs(output_path)

    # Update the fine-tuning configuration
    fine_tuning_cfg = manage_fine_tuning_config(output_path, num_epochs, batch_size,freeze_layers)

    # Set the path of the pre-trained model in the configuration
    fine_tuning_cfg["saved_model"]["saved_model_path"] = pretrained_model_path
    fine_tuning_cfg["saved_model"]["use_saved_model"] = True

    run_main_train_code(fine_tuning_cfg)


def manage_fine_tuning_config(output_path, num_epochs, batch_size, freeze_layers):
    # Define the paths to the source and destination JSON files
    working_dir = os.path.realpath(os.path.dirname(__file__))
    config_base_path = os.path.join(working_dir, "ramp_config_base.json")
    dst_path = os.path.join(output_path, "ramp_fair_config_finetune.json")

    # Read the content of the source JSON file
    with open(config_base_path, "r") as f:
        data = json.load(f)

    # Modify the content of the data dictionary datasets
    data["datasets"]["train_img_dir"] = f"{output_path}/chips"
    data["datasets"]["train_mask_dir"] = f"{output_path}/binarymasks"
    data["datasets"]["val_img_dir"] = f"{output_path}/val-chips"
    data["datasets"]["val_mask_dir"] = f"{output_path}/val-binarymasks"

    # epoch batchconfig
    data["num_epochs"] = num_epochs
    data["batch_size"] = batch_size
    data["freeze_layers"] = freeze_layers

    # clr plot
    data["cyclic_learning_scheduler"]["clr_plot_dir"] = f"{output_path}/plots"
    # logs
    data["tensorboard"]["tb_logs_dir"] = f"{output_path}/logs"
    # output images
    data["graph_location"] = f"{output_path}/graphs"
    # model_checkpts
    data["model_checkpts"]["model_checkpts_dir"] = f"{output_path}/model-checkpts"
    # save best models only
    data["model_checkpts"]["model_checkpt_callback_parms"]["save_best_only"] = True
    # Open the destination file and write the modified data
    with open(dst_path, "w") as f:
        json.dump(data, f)

    return data


def run_main_train_code(cfg):
    discard_experiment = False
    if "discard_experiment" in cfg:
        discard_experiment = cfg["discard_experiment"]

    cfg["timestamp"] = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")

    # specify a function that will construct the loss function
    get_loss_fn_name = cfg["loss"]["get_loss_fn_name"]
    get_loss_fn = getattr(loss_constructors, get_loss_fn_name)

    # Construct the loss function
    loss_fn = get_loss_fn(cfg)

    the_metrics = []
    if cfg["metrics"]["use_metrics"]:
        get_metrics_fn_names = cfg["metrics"]["get_metrics_fn_names"]
        get_metrics_fn_parms = cfg["metrics"]["metrics_fn_parms"]

        for get_mf_name, mf_parms in zip(get_metrics_fn_names, get_metrics_fn_parms):
            get_metric_fn = getattr(metric_constructors, get_mf_name)
            print(f"Metric constructor function: {get_metric_fn.__name__}")
            metric_fn = get_metric_fn(mf_parms)
            the_metrics.append(metric_fn)

    #### construct optimizer ####

    get_optimizer_fn_name = cfg["optimizer"]["get_optimizer_fn_name"]
    get_optimizer_fn = getattr(optimizer_constructors, get_optimizer_fn_name)

    optimizer = get_optimizer_fn(cfg)

    the_model = None

    if cfg["saved_model"]["use_saved_model"]:
        # load (construct) the model
        model_path = Path(working_ramp_home) / cfg["saved_model"]["saved_model_path"]
        print(f"Model: importing saved model {str(model_path)}")
        the_model = tf.keras.models.load_model(model_path)
        assert (
            the_model is not None
        ), f"the saved model was not constructed: {model_path}"

        if cfg["freeze_layers"]:
            num_layers_to_freeze = 4 # freeze lower layers 
            for index, layer in enumerate(the_model.layers):
                if index < num_layers_to_freeze:
                    layer.trainable = False
                else:
                    layer.trainable = True


        if not cfg["saved_model"]["save_optimizer_state"]:
            # If you don't want to save the original state of training, recompile the model.
            the_model.compile(optimizer=optimizer, loss=loss_fn, metrics=[the_metrics])

            # the_model.compile(optimizer = optimizer,
            #    loss=loss_fn,
            #    metrics = [get_iou_coef_fn])

    if not cfg["saved_model"]["use_saved_model"]:
        get_model_fn_name = cfg["model"]["get_model_fn_name"]
        get_model_fn = getattr(model_constructors, get_model_fn_name)
        the_model = get_model_fn(cfg)

        assert the_model is not None, f"the model was not constructed: {model_path}"
        the_model.compile(optimizer=optimizer, loss=loss_fn, metrics=the_metrics)

    print(the_model)
    cfg["datasets"]

    #### define data directories ####
    train_img_dir = Path(working_ramp_home) / cfg["datasets"]["train_img_dir"]
    train_mask_dir = Path(working_ramp_home) / cfg["datasets"]["train_mask_dir"]
    val_img_dir = Path(working_ramp_home) / cfg["datasets"]["val_img_dir"]
    val_mask_dir = Path(working_ramp_home) / cfg["datasets"]["val_mask_dir"]

    #### get the augmentation transform ####
    # aug = None
    if cfg["augmentation"]["use_aug"]:
        aug = get_augmentation_fn(cfg)

    ## RUNTIME Parameters
    batch_size = cfg["batch_size"]
    input_img_shape = cfg["input_img_shape"]
    output_img_shape = cfg["output_img_shape"]

    n_training = get_num_files(train_img_dir, "*.tif")
    n_val = get_num_files(val_img_dir, "*.tif")
    steps_per_epoch = n_training // batch_size
    validation_steps = n_val // batch_size
    # Testing step , not recommended
    if validation_steps <= 0:
        validation_steps = 1

    # add these back to the config
    # in case they are needed by callbacks
    cfg["runtime"] = {}
    cfg["runtime"]["n_training"] = n_training
    cfg["runtime"]["n_val"] = n_val
    cfg["runtime"]["steps_per_epoch"] = steps_per_epoch
    cfg["runtime"]["validation_steps"] = validation_steps

    train_batches = None

    if aug is not None:
        train_batches = training_batches_from_gtiff_dirs(
            train_img_dir,
            train_mask_dir,
            batch_size,
            input_img_shape,
            output_img_shape,
            transforms=aug,
        )
    else:
        train_batches = training_batches_from_gtiff_dirs(
            train_img_dir, train_mask_dir, batch_size, input_img_shape, output_img_shape
        )

    assert train_batches is not None, "training batches were not constructed"

    val_batches = test_batches_from_gtiff_dirs(
        val_img_dir, val_mask_dir, batch_size, input_img_shape, output_img_shape
    )

    assert val_batches is not None, "validation batches were not constructed"

    ## Callbacks ##
    callbacks_list = []

    if not discard_experiment:
        # get model checkpoint callback
        if cfg["model_checkpts"]["use_model_checkpts"]:
            get_model_checkpt_callback_fn_name = cfg["model_checkpts"][
                "get_model_checkpt_callback_fn_name"
            ]
            get_model_checkpt_callback_fn = getattr(
                callback_constructors, get_model_checkpt_callback_fn_name
            )
            callbacks_list.append(get_model_checkpt_callback_fn(cfg))

        # get tensorboard callback
        if cfg["tensorboard"]["use_tb"]:
            get_tb_callback_fn_name = cfg["tensorboard"]["get_tb_callback_fn_name"]
            get_tb_callback_fn = getattr(callback_constructors, get_tb_callback_fn_name)
            callbacks_list.append(get_tb_callback_fn(cfg))

        # get tensorboard model prediction logging callback
        if cfg["prediction_logging"]["use_prediction_logging"]:
            assert cfg["tensorboard"][
                "use_tb"
            ], "Tensorboard logging must be turned on to enable prediction logging"
            get_prediction_logging_fn_name = cfg["prediction_logging"][
                "get_prediction_logging_fn_name"
            ]
            get_prediction_logging_fn = getattr(
                callback_constructors, get_prediction_logging_fn_name
            )
            callbacks_list.append(get_prediction_logging_fn(the_model, cfg))

    # free up RAM
    keras.backend.clear_session()

    if cfg["early_stopping"]["use_early_stopping"]:
        callbacks_list.append(callback_constructors.get_early_stopping_callback_fn(cfg))

        # get cyclic learning scheduler callback
    if cfg["cyclic_learning_scheduler"]["use_clr"]:
        assert not cfg["early_stopping"][
            "use_early_stopping"
        ], "cannot use early_stopping with cycling_learning_scheduler"
        get_clr_callback_fn_name = cfg["cyclic_learning_scheduler"][
            "get_clr_callback_fn_name"
        ]
        get_clr_callback_fn = getattr(callback_constructors, get_clr_callback_fn_name)
        callbacks_list.append(get_clr_callback_fn(cfg))

    ## Main training block ##
    n_epochs = cfg["num_epochs"]
    print(
        f"Starting Training with {n_epochs} epochs , {batch_size} batch size , {steps_per_epoch} steps per epoch , {validation_steps} validation steps......"
    )
    if validation_steps <= 0:
        raise RaiseError(
            "Not enough data for training, Increase image or Try reducing batchsize/epochs"
        )
    # FIXME : Make checkpoint
    start = perf_counter()
    history = the_model.fit(
        train_batches,
        epochs=n_epochs,
        steps_per_epoch=steps_per_epoch,
        validation_data=val_batches,
        validation_steps=validation_steps,
        callbacks=callbacks_list,
    )
    end = perf_counter()
    print(f"Training Finished , Time taken to train : {end-start} seconds")

    # plot the training and validation accuracy and loss at each epoch
    print("Generating graphs ....")
    if not os.path.exists(cfg["graph_location"]):
        os.mkdir(cfg["graph_location"])

    loss = history.history["loss"]
    # val_loss = history.history["val_loss"]
    epochs = range(1, len(loss) + 1)

    acc = history.history["sparse_categorical_accuracy"]
    val_acc = history.history["val_sparse_categorical_accuracy"]

    plt.figure()

    # Plot training and validation accuracy
    plt.plot(epochs, acc, "y", label="Training Accuracy")
    plt.plot(epochs, val_acc, "r", label="Validation Accuracy")

    # Set labels and title
    plt.xlabel("Epochs")
    plt.ylabel("Accuracy")
    plt.title("Training and Validation Accuracy")

    plt.legend()
    plt.savefig(
        f"{cfg['graph_location']}/training_validation_sparse_categorical_accuracy.png"
    )
    plt.clf()
    print(f"Graph generated at : {cfg['graph_location']}")



# Third-party imports
import geopandas


def reproject_labels_to_epsg3857(input_path: str, output_path: str) -> None:
    """Convert a GeoJSON file with labels from EPSG:4326 to EPSG:3857.

    A new GeoJSON file is created, it contains coordinates in meters
    (easting, northing) in the 'WGS 84 / Pseudo-Mercator' projection.

    Args:
        input_path: Path to the GeoJSON file where the input data are stored.
        output_path: Path to the GeoJSON file where the output data will go.
    """
    labels_gdf = geopandas.read_file(input_path).set_crs("EPSG:4326")
    labels_gdf.to_crs("EPSG:3857").to_file(output_path)



# Standard library imports
import os
from pathlib import Path

from .get_polygons import get_polygons
from .merge_polygons import merge_polygons


def polygonize(input_path: str, output_path: str, remove_inputs: False) -> None:
    """Polygonize raster tiles from the input path using AutoBFE's algorithm.

    There are two steps:
    1. It polygonizes each of the individual tiles first.
    2. Using GDAL buffering, it merges all the nearby polygons
    so that the split polygons get merged into one.

    Whether the images are georeferenced or not doesn't really matter.
    Even PNG images will do.

    CRS of the resulting GeoJSON file will be EPSG:4326.

    Args:
        input_path: Path of the directory where the image files are stored.
        output_path: Path of the output file.
        remove_inputs: Clears the input image after geojson is produced

    Example::

        polygonize("data/masks_v2/4", "labels.geojson")
    """
    if os.path.exists(output_path):
        os.remove(output_path)
    base_path = Path(output_path).parents[0]
    base_path.mkdir(exist_ok=True, parents=True)

    get_polygons(input_path, "temp-labels.geojson", kernel_opening=1)
    merge_polygons("temp-labels.geojson", output_path, distance_threshold=0.6)
    os.remove("temp-labels.geojson")

    if remove_inputs:
        # Get a list of all .tif files in the directory
        tif_files = [f for f in os.listdir(input_path) if f.endswith(".tif")]

        # Loop through the list of .tif files and delete each one
        for file in tif_files:
            file_path = os.path.join(input_path, file)
            os.remove(file_path)



# Third party imports
import geopandas
from shapely.validation import explain_validity, make_valid


def remove_self_intersection(row):
    """Fix self-intersections in the polygons.

    Some of the polygons may have self-intersections. In that
    case, we transform that geometry to a multi-polygon and
    substitute the original geometry with the largest polygon.
    """
    if explain_validity(row.geometry) == "Valid Geometry":
        return row.geometry

    valid_geom = make_valid(row.geometry)
    if not hasattr(valid_geom, "__len__"):
        return valid_geom

    for polygon in valid_geom.geoms:
        if polygon.area >= row.geometry.area / 2.0:
            return polygon


def fix_labels(input_path: str, output_path: str) -> None:
    """Fix GeoJSON file so that it doesn't have any self-intersecting polygons.

    Args:
        input_path: Path to the GeoJSON file where the input data are stored.
        output_path: Path to the GeoJSON file where the output data will go.
    """
    gdf = geopandas.read_file(input_path)
    if gdf.empty:
        raise ValueError("Error: gdf is empty, No Labels found : Check your labels")
    gdf["geometry"] = gdf.apply(remove_self_intersection, axis=1)
    gdf.to_file(output_path)



import collections

from geopandas import GeoSeries
from shapely.geometry import MultiPolygon, Polygon
from shapely.ops import unary_union
from shapely.validation import make_valid

from .utils import extract_contours, featurize, opening, parents_in_hierarchy, simplify

CRS = "EPSG:4326"


class BuildingExtract(object):
    def __init__(self, kernel_opening=20, simplify_threshold=0.01):
        """Adapted from: https://github.com/mapbox/robosat"""
        self.kernel_opening = kernel_opening
        self.simplify_threshold = simplify_threshold
        self.features = []

    def extract(self, tile, mask):
        mask = opening(mask, self.kernel_opening)
        multipolygons, hierarchy = extract_contours(mask)

        if hierarchy is None:
            return

        assert (
            len(hierarchy) == 1
        ), "always single hierarchy for all polygons in multipolygon"
        hierarchy = hierarchy[0]

        assert len(multipolygons) == len(hierarchy), "polygons and hierarchy in sync"

        polygons = [
            simplify(polygon, self.simplify_threshold) for polygon in multipolygons
        ]

        # All child ids in hierarchy tree, keyed by root id.
        features = collections.defaultdict(set)

        for i, (polygon, _) in enumerate(zip(polygons, hierarchy)):
            if len(polygon) < 3:
                continue

            ancestors = list(parents_in_hierarchy(i, hierarchy))

            # Only handles polygons with a nesting of two levels for now => no multipolygons.
            if len(ancestors) > 1:
                continue

            # A single mapping: i => {i} implies single free-standing polygon, no inner rings.
            # Otherwise: i => {i, j, k, l} implies: outer ring i, inner rings j, k, l.
            root = ancestors[-1] if ancestors else i

            features[root].add(i)

        for outer, inner in features.items():
            rings = [featurize(tile, polygons[outer], mask.shape[:2])]

            # In mapping i => {i, ..} i is not a child.
            children = inner.difference(set([outer]))

            for child in children:
                rings.append(featurize(tile, polygons[child], mask.shape[:2]))

            feature = make_valid(
                unary_union([make_valid(Polygon(ring)) for ring in rings])
            )

            if type(feature) == MultiPolygon:
                for polygon in feature.geoms:
                    if type(polygon) == Polygon and polygon.area > 0:
                        self.features.append(polygon)
            elif type(feature) == Polygon:
                self.features.append(feature)

    def save(self, out):
        GeoSeries(self.features).set_crs(CRS).to_file(out)



# Third party imports
from geopandas import GeoSeries, read_file
from shapely.geometry import MultiPolygon, Polygon
from shapely.validation import make_valid
from tqdm import tqdm

from .utils import UndirectedGraph, make_index, project, union

TOLERANCE = 1e-6
SOURCE_CRS = "EPSG:4326"
INTERMEDIATE_CRS = "EPSG:3395"


def merge_polygons(polygons_path, new_polygons_path, distance_threshold):
    """Merge polygons from GeoJSON file.

    Args:
      polygons_path: GeoJSON file to read polygons from
      new_polygons_path: Path to GeoJSON file where the merged polygons will be saved
      distance_threshold: Minimum distance to define adjacent polygons, in meters
    """
    gdf = read_file(polygons_path)
    shapes = list(gdf["geometry"])

    graph = UndirectedGraph()
    idx = make_index(shapes)

    def buffered(shape):
        projected = project(shape, SOURCE_CRS, INTERMEDIATE_CRS)
        buffered = projected.buffer(distance_threshold)
        unprojected = project(buffered, INTERMEDIATE_CRS, SOURCE_CRS)

        return unprojected

    def unbuffered(shape):
        projected = project(shape, SOURCE_CRS, INTERMEDIATE_CRS)
        unbuffered = projected.buffer(-1 * distance_threshold)
        unprojected = project(unbuffered, INTERMEDIATE_CRS, SOURCE_CRS)

        return unprojected

    for i, shape in enumerate(tqdm(shapes, desc="Building graph", unit="shapes")):
        embiggened = buffered(shape)
        graph.add_edge(i, i)
        nearest = [
            j for j in idx.intersection(embiggened.bounds, objects=False) if i != j
        ]

        for t in nearest:
            if embiggened.intersects(shapes[t]):
                graph.add_edge(i, t)

    components = list(graph.components())
    assert sum([len(v) for v in components]) == len(
        shapes
    ), "components capture all shape indices"

    features = []

    for component in tqdm(components, desc="Merging components", unit="component"):
        embiggened = [buffered(shapes[v]) for v in component]
        merged = unbuffered(union(embiggened))

        feature = make_valid(merged)

        if type(feature) == MultiPolygon:
            for polygon in feature.geoms:
                if type(polygon) == Polygon and polygon.area > 0:
                    features.append(polygon)
        elif type(feature) == Polygon:
            features.append(feature)

    gs = GeoSeries(features).set_crs(SOURCE_CRS)
    gs.simplify(TOLERANCE).to_file(new_polygons_path)



from typing import List

import numpy as np
from PIL import Image
from tensorflow import keras

IMAGE_SIZE = 256


def open_images(paths: List[str]) -> np.ndarray:
    """Open images from some given paths."""
    images = []
    for path in paths:
        image = keras.preprocessing.image.load_img(
            path, target_size=(IMAGE_SIZE, IMAGE_SIZE)
        )
        image = np.array(image.getdata()).reshape(IMAGE_SIZE, IMAGE_SIZE, 3) / 255.0
        images.append(image)

    return np.array(images)


def save_mask(mask: np.ndarray, filename: str) -> None:
    """Save the mask array to the specified location."""
    reshaped_mask = mask.reshape((IMAGE_SIZE, IMAGE_SIZE)) * 255
    result = Image.fromarray(reshaped_mask.astype(np.uint8))
    result.save(filename)



# Standard library imports
import os
import shutil


def extract_highest_accuracy_model(output_path):
    model_checkpoints = os.path.join(output_path, "model-checkpts")
    assert os.path.exists(model_checkpoints), "Model Checkpoints Doesn't Exist"
    entries = os.listdir(model_checkpoints)
    assert len(entries) > 0, "Couldn't find any models"
    # Create a variable to store the latest entry
    latest_entry = None

    # Create a variable to store the latest entry's modification time
    latest_time = 0

    # Iterate over the list of entries
    for entry in entries:
        # Use the os.stat() method to get the entry's modification time
        entry_time = os.stat(os.path.join(model_checkpoints, entry)).st_mtime
        # Check if the entry's modification time is greater than the latest time
        if entry_time > latest_time:
            # If the entry's modification time is greater, update the latest time and entry variables
            latest_time = entry_time
            latest_entry = entry

    # get the highest accuracy model one
    latest_entry_path = os.path.join(model_checkpoints, latest_entry)
    print(latest_entry_path)
    highest_accuracy = 0
    highest_accuracy_entry = None
    for entry in os.listdir(latest_entry_path):

        parts = entry.split("_")

        accuracy = parts[-1][:-3]  # remove .tf
        if float(accuracy) * 100 > highest_accuracy:
            highest_accuracy_entry = entry
            highest_accuracy = float(accuracy) * 100
    print(highest_accuracy_entry)
    for entry in os.listdir(latest_entry_path):
        # Check if the entry is not the file or directory you want to keep
        if entry != highest_accuracy_entry:
            # If the entry is not the file or directory you want to keep, use the os.remove() method to remove it
            shutil.rmtree(os.path.join(latest_entry_path, entry))
    return highest_accuracy, os.path.join(latest_entry_path, highest_accuracy_entry)



# Standard library imports
import os
from glob import glob
from pathlib import Path

# Third party imports
# Third-party imports
from osgeo import gdal
from tqdm import tqdm

from .utils import get_bounding_box


def georeference(input_path: str, output_path: str, is_mask=False) -> None:
    """Perform georeferencing and remove the fourth band from images (if any).

    CRS of the georeferenced images will be EPSG:3857 ('WGS 84 / Pseudo-Mercator').

    Args:
        input_path: Path of the directory where the input data are stored.
        output_path: Path of the directory where the output data will go.
        is_mask: Whether the image is binary or not.

    Example::

        georeference(
            "data/prediction-dataset/5x5/1-19",
            "data/georeferenced_input/1-19"
        )
    """
    os.makedirs(output_path, exist_ok=True)

    for path in tqdm(
        glob(f"{input_path}/*.png"), desc=f"Georeferencing for {Path(input_path).stem}"
    ):
        filename = Path(path).stem
        in_file = f"{input_path}/{filename}.png"
        out_file = f"{output_path}/{filename}.tif"

        # Get bounding box in EPSG:3857
        x_min, y_min, x_max, y_max = get_bounding_box(filename)

        # Use one band for masks and the first three bands for images
        bands = [1] if is_mask else [1, 2, 3]

        # Georeference image
        # Output bounds are defined by upper left and lower right corners
        _ = gdal.Translate(
            destName=out_file,
            srcDS=in_file,
            format="GTiff",
            bandList=bands,
            outputBounds=[x_min, y_max, x_max, y_min],
            outputSRS="EPSG:3857",
        )
        # Close dataset
        _ = None



import collections
import functools
import re
import warnings
from glob import glob
from pathlib import Path

import cv2
import mercantile
import pyproj
import shapely.ops
from rtree.index import Index, Property


def tiles_from_directory(dir_path):
    """Loads files from a directory
    Adapted from: https://github.com/mapbox/robosat

    Args:
      root: the  directory with files that have a name with OAM structure "OAM-x-y-z.*""

    Yields:
      The mercantile tiles and file paths
    """
    for path in glob(f"{dir_path}/*"):
        _, *tile_info = re.split("-", Path(path).stem)
        x, y, z = map(int, tile_info)
        tile = mercantile.Tile(x=x, y=y, z=z)
        yield tile, path


def opening(mask, kernel_size):
    """Morphologycal transforations of erosion the dilatation which removes small objects
    Args:
      mask: the binary mask to transform
      eps: the opening kernel size, in pixel

    Returns:
      The transformed mask
    """
    struct = cv2.getStructuringElement(cv2.MORPH_RECT, (kernel_size, kernel_size))
    return cv2.morphologyEx(mask, cv2.MORPH_OPEN, struct)


def extract_contours(mask):
    """Find contours objects in the mask
    From: https://github.com/mapbox/robosat
    Args:
      mask: the binary mask to transform
      eps: the opening kernel size, in pixel

    Returns:
      The detected contours as a list of points and the contour hierarchy
    """
    contours, hierarchy = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
    return contours, hierarchy


def simplify(polygon, eps):
    """Simplifies a polygon to minimize the polygon's vertices
    From: https://github.com/mapbox/robosat

    Args:
      polygon: the polygon made up of a list of vertices.
      eps: the approximation accuracy as max. percentage of the arc length, in [0, 1]
    Returns:
      The simplified polygon
    """
    assert 0 <= eps <= 1, "approximation accuracy is percentage in [0, 1]"

    epsilon = eps * cv2.arcLength(polygon, closed=True)
    return cv2.approxPolyDP(polygon, epsilon=epsilon, closed=True)


def parent(n):
    # next, prev, fst child, parent
    return n[3]


def parents_in_hierarchy(node, tree):
    """Walks a hierarchy tree upwards from a starting node collecting all nodes on the way.
    From: https://github.com/mapbox/robosat

    Args:
      node: the index for the starting node in the hierarchy.
      tree: the hierarchy tree containing tuples of (next, prev, first child, parent) ids.

    Yields:
      The node ids on the upwards path in the hierarchy tree.
    """
    at = tree[node]
    up = parent(at)

    while up != -1:
        index = up
        at = tree[index]
        up = parent(at)

        assert index != node, "upward path does not include starting node"

        yield index


def featurize(tile, polygon, shape):
    """Transforms polygons in image pixel coordinates into world coordinates.
    From: https://github.com/mapbox/robosat

    Args:
      tile: the tile this polygon is in for coordinate calculation.
      polygon: the polygon to transform from pixel to world coordinates.
      shape: the image's max x and y coordinates.

    Returns:
      The closed polygon transformed into world coordinates.
    """
    xmax, ymax = shape

    feature = []

    for point in polygon:
        px, py = point[0]
        dx, dy = px / xmax, py / ymax

        feature.append(pixel_to_location(tile, dx, 1.0 - dy))

    assert feature, "at least one location in polygon"
    feature.append(feature[0])  # polygons are closed

    return feature


def pixel_to_location(tile, dx, dy):
    """Converts a pixel in a tile to a coordinate.

    Args:
      tile: the mercantile tile to calculate the location in.
      dx: the relative x offset in range [0, 1].
      dy: the relative y offset in range [0, 1].

    Returns:
      The coordinate for the pixel in the tile.
    """
    assert 0 <= dx <= 1, "x offset is in [0, 1]"
    assert 0 <= dy <= 1, "y offset is in [0, 1]"

    west, south, east, north = mercantile.bounds(tile)

    def lerp(a, b, c):
        return a + c * (b - a)

    lon = lerp(west, east, dx)
    lat = lerp(south, north, dy)

    return lon, lat


class UndirectedGraph:
    """Simple undirected graph.
    From: https://github.com/mapbox/robosat

    Note: stores edges; can not store vertices without edges.
    """

    def __init__(self):
        """Creates an empty `UndirectedGraph` instance."""

        # Todo: We might need a compressed sparse row graph (i.e. adjacency array)
        # to make this scale. Let's circle back when we run into this limitation.
        self.edges = collections.defaultdict(set)

    def add_edge(self, s, t):
        """Adds an edge to the graph.

        Args:
          s: the source vertex.
          t: the target vertex.

        Note: because this is an undirected graph for every edge `s, t` an edge `t, s` is added.
        """
        self.edges[s].add(t)
        self.edges[t].add(s)

    def targets(self, v):
        """Returns all outgoing targets for a vertex.

        Args:
          v: the vertex to return targets for.

        Returns:
          A list of all outgoing targets for the vertex.
        """
        return self.edges[v]

    def vertices(self):
        """Returns all vertices in the graph.

        Returns:
          A set of all vertices in the graph.
        """
        return self.edges.keys()

    def empty(self):
        """Returns true if the graph is empty, false otherwise.

        Returns:
          True if the graph has no edges or vertices, false otherwise.
        """
        return len(self.edges) == 0

    def dfs(self, v):
        """Applies a depth-first search to the graph.

        Args:
          v: the vertex to start the depth-first search at.

        Yields:
          The visited graph vertices in depth-first search order.

        Note: does not include the start vertex `v` (except if an edge targets it).
        """
        stack = []
        stack.append(v)

        seen = set()

        while stack:
            s = stack.pop()

            if s not in seen:
                seen.add(s)

                for t in self.targets(s):
                    stack.append(t)

                yield s

    def components(self):
        """Computes connected components for the graph.

        Yields:
          The connected component sub-graphs consisting of vertices; in no particular order.
        """
        seen = set()

        for v in self.vertices():
            if v not in seen:
                component = set(self.dfs(v))
                component.add(v)

                seen.update(component)

                yield component


def project(shape, source, target):
    """Projects a geometry from one coordinate system into another.
    This function is an adaptation to bypass a bug in pyproj package
    Args:
      shape: the geometry to project.
      source: the source EPSG spatial reference system identifier.
      target: the target EPSG spatial reference system identifier.

    Returns:
      The projected geometry in the target coordinate system.
    """
    with warnings.catch_warnings():  # To exclude the warnings of Proj deprecation
        warnings.simplefilter("ignore")
        proj_in = pyproj.Proj(init=source)
        proj_out = pyproj.Proj(init=target)
    project_fun = pyproj.Transformer.from_proj(proj_in, proj_out).transform

    return shapely.ops.transform(project_fun, shape)


def union(shapes):
    """Returns the union of all shapes.
    From: https://github.com/mapbox/robosat

    Args:
      shapes: the geometries to merge into one.

    Returns:
      The union of all shapes as one shape.
    """
    assert shapes

    def fn(lhs, rhs):
        return lhs.union(rhs)

    return functools.reduce(fn, shapes)


def make_index(shapes):
    """Creates an index for fast and efficient spatial queries.
    From: https://github.com/mapbox/robosat

    Args:
      shapes: shapely shapes to bulk-insert bounding boxes for into the spatial index.

    Returns:
      The spatial index created from the shape's bounding boxes.
    """
    # Todo: benchmark these for our use-cases
    prop = Property()
    prop.dimension = 2
    prop.leaf_capacity = 1000
    prop.fill_factor = 0.9

    def bounded():
        for i, shape in enumerate(shapes):
            yield (i, shape.bounds, None)

    return Index(bounded(), properties=prop)
